#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <libguile.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <nettle/rsa.h>

#include <stdio.h>

struct nettle_random_func_box
{
  void *ctx;
  void (*f) (void *, size_t, uint8_t *);
};

static void the_nettle_random_func (SCM * func,
                                    struct nettle_random_func_box *f);

static SCM rsa_public_key_type;
static SCM rsa_private_key_type;
static SCM stubs_with_rsa_generated_key_pair (SCM rng, SCM progress,
                                              SCM n_size, SCM e_size, SCM f);
static struct rsa_public_key *the_rsa_public_key (SCM key);

static SCM stubs_rsa_public_key_n (SCM key);

static void
guile_nettle_random_func (void *ctx, size_t length, uint8_t * dst)
{
  SCM *cb = ctx;
  SCM scm_length = scm_from_size_t (length);
  SCM random_data = scm_call_1 (*cb, scm_length);
  size_t output_length = scm_c_bytevector_length (random_data);
  size_t i;
  SCM_ASSERT (output_length == length,
              *cb, SCM_ARGn, "generating random data");
  for (i = 0; i < length; i++)
    {
      dst[i] = scm_c_bytevector_ref (random_data, i);
    }
}

static void
the_nettle_random_func (SCM * func, struct nettle_random_func_box *f)
{
  f->ctx = func;
  f->f = guile_nettle_random_func;
}

static void
do_rsa_public_key_clear (void *data)
{
  struct rsa_public_key *pub = data;
  rsa_public_key_clear (pub);
}

static void
do_rsa_private_key_clear (void *data)
{
  struct rsa_private_key *key = data;
  rsa_private_key_clear (key);
}

static SCM
stubs_with_rsa_generated_key_pair (SCM rng, SCM progress, SCM n_size,
                                   SCM e_size, SCM f)
{
  struct rsa_public_key *c_pub =
    scm_gc_malloc_pointerless (sizeof (struct rsa_public_key),
                               "rsa-public-key");
  struct rsa_private_key *c_key =
    scm_gc_malloc_pointerless (sizeof (struct rsa_private_key),
                               "rsa-private-key");
  struct nettle_random_func_box c_random;
  unsigned c_n_size = scm_to_uint (n_size);
  unsigned c_e_size = scm_to_uint (e_size);
  int ok;
  SCM pub, key, ret;
  the_nettle_random_func (&rng, &c_random);
  (void) progress;
  scm_dynwind_begin (0);
  rsa_public_key_init (c_pub);
  scm_dynwind_unwind_handler (do_rsa_public_key_clear, c_pub,
                              SCM_F_WIND_EXPLICITLY);
  rsa_private_key_init (c_key);
  scm_dynwind_unwind_handler (do_rsa_private_key_clear, c_key,
                              SCM_F_WIND_EXPLICITLY);
  ok =
    rsa_generate_keypair (c_pub, c_key, c_random.ctx, c_random.f,
                          NULL, NULL, c_n_size, c_e_size);
  if (ok)
    {
      pub = scm_make_foreign_object_1 (rsa_public_key_type, c_pub);
      key = scm_make_foreign_object_1 (rsa_private_key_type, c_key);
      ret = scm_call_2 (f, pub, key);
    }
  else
    {
      ret = scm_call_2 (f, SCM_BOOL_F, SCM_BOOL_F);
    }
  scm_dynwind_end ();
  return ret;
}

static struct rsa_public_key *
the_rsa_public_key (SCM key)
{
  scm_assert_foreign_object_type (rsa_public_key_type, key);
  return scm_foreign_object_ref (key, 0);
}

static SCM
stubs_rsa_public_key_n (SCM key)
{
  return scm_from_mpz (the_rsa_public_key (key)->n);
}

void
init_nettle (void)
{
  {
    SCM name = scm_from_utf8_symbol ("rsa-public-key");
    SCM slots = scm_list_1 (scm_from_utf8_symbol ("ptr"));
    scm_t_struct_finalize finalizer = NULL;
    rsa_public_key_type =
      scm_make_foreign_object_type (name, slots, finalizer);
    scm_c_define_gsubr ("rsa-public-key-n", 1, 0, 0,
                        stubs_rsa_public_key_n);
  }
  {
    SCM name = scm_from_utf8_symbol ("rsa-private-key");
    SCM slots = scm_list_1 (scm_from_utf8_symbol ("ptr"));
    scm_t_struct_finalize finalizer = NULL;
    rsa_private_key_type =
      scm_make_foreign_object_type (name, slots, finalizer);
  }
  scm_c_define_gsubr ("with-rsa-generated-key-pair", 5, 0, 0,
                      stubs_with_rsa_generated_key_pair);
}
